@@ -67,6 +67,8 @@ Please remember to secure your keys and access tokens.
 
 from __future__ import with_statement
 
+from typing import Optional
+
 try:
     import simplejson as json
 except ImportError:
@@ -83,22 +85,21 @@ import sys
 
 from libcloud.utils.connection import get_response_object
 from libcloud.utils.py3 import b, httplib, urlencode, urlparse, PY3
-from libcloud.common.base import (ConnectionUserAndKey, JsonResponse,
-                                  PollingConnection)
+from libcloud.common.base import ConnectionUserAndKey, JsonResponse, PollingConnection
 from libcloud.common.base import BaseDriver
-from libcloud.common.types import (ProviderError,
-                                   LibcloudError)
+from libcloud.common.types import ProviderError, LibcloudError
 
 try:
     from cryptography.hazmat.backends import default_backend
     from cryptography.hazmat.primitives import serialization
     from cryptography.hazmat.primitives.hashes import SHA256
     from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
+    from cryptography import exceptions
 except ImportError:
     # The cryptography library is unavailable
     SHA256 = None  # type: ignore
 
-UTC_TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
+UTC_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
 
 LOG = logging.getLogger(__name__)
 
@@ -124,18 +125,19 @@ def _from_utc_timestamp(timestamp):
     return datetime.datetime.strptime(timestamp, UTC_TIMESTAMP_FORMAT)
 
 
-def _get_gce_metadata(path=''):
+def _get_gce_metadata(path="", retry_failed: Optional[bool] = None):
     try:
-        url = 'http://metadata/computeMetadata/v1/' + path.lstrip('/')
-        headers = {'Metadata-Flavor': 'Google'}
-        response = get_response_object(url, headers=headers)
-        return response.status, '', response.body
+        url = "http://metadata/computeMetadata/v1/" + path.lstrip("/")
+        headers = {"Metadata-Flavor": "Google"}
+        response = get_response_object(url, headers=headers, retry_failed=retry_failed)
+        return response.status, "", response.body
     except Exception as e:
         return -1, str(e), None
 
 
 class GoogleAuthError(LibcloudError):
     """Generic Error class for various authentication errors."""
+
     def __init__(self, value):
         self.value = value
 
@@ -160,13 +162,18 @@ class JsonParseError(GoogleBaseError):
 class ResourceNotFoundError(GoogleBaseError):
     def __init__(self, value, http_code, code, driver=None):
         self.code = code
-        if isinstance(value, dict) and 'message' in value and \
-                value['message'].count('/') == 1 and \
-                value['message'].count('projects/') == 1:
-            value['message'] = value['message'] + ". A missing project " \
-                "error may be an authentication issue. " \
-                "Please  ensure your auth credentials match " \
+        if (
+            isinstance(value, dict)
+            and "message" in value
+            and value["message"].count("/") == 1
+            and value["message"].count("projects/") == 1
+        ):
+            value["message"] = (
+                value["message"] + ". A missing project "
+                "error may be an authentication issue. "
+                "Please  ensure your auth credentials match "
                 "your project. "
+            )
         super(ResourceNotFoundError, self).__init__(value, http_code, driver)
 
 
@@ -186,6 +193,7 @@ class GoogleResponse(JsonResponse):
     """
     Google Base Response class.
     """
+
     def success(self):
         """
         Determine if the request was successful.
@@ -209,19 +217,19 @@ class GoogleResponse(JsonResponse):
         :return:  Tuple containing error code and message
         :rtype:   ``tuple`` of ``str`` or ``int``
         """
-        if 'errors' in body['error']:
-            err = body['error']['errors'][0]
+        if "errors" in body["error"]:
+            err = body["error"]["errors"][0]
         else:
-            err = body['error']
+            err = body["error"]
 
-        if 'code' in err:
-            code = err.get('code')
-            message = err.get('message')
+        if "code" in err:
+            code = err.get("code")
+            message = err.get("message")
         else:
             code = None
-            if 'reason' in err:
-                code = err.get('reason')
-            message = body.get('error_description', err)
+            if "reason" in err:
+                code = err.get("reason")
+            message = body.get("error_description", err)
 
         return (code, message)
 
@@ -254,15 +262,15 @@ class GoogleResponse(JsonResponse):
         if self.status in valid_http_codes:
             if json_error:
                 raise JsonParseError(body, self.status, None)
-            elif 'error' in body:
+            elif "error" in body:
                 (code, message) = self._get_error(body)
-                if code == 'QUOTA_EXCEEDED':
+                if code == "QUOTA_EXCEEDED":
                     raise QuotaExceededError(message, self.status, code)
-                elif code == 'RESOURCE_ALREADY_EXISTS':
+                elif code == "RESOURCE_ALREADY_EXISTS":
                     raise ResourceExistsError(message, self.status, code)
-                elif code == 'alreadyExists':
+                elif code == "alreadyExists":
                     raise ResourceExistsError(message, self.status, code)
-                elif code.startswith('RESOURCE_IN_USE'):
+                elif code.startswith("RESOURCE_IN_USE"):
                     raise ResourceInUseError(message, self.status, code)
                 else:
                     raise GoogleBaseError(message, self.status, code)
@@ -270,7 +278,7 @@ class GoogleResponse(JsonResponse):
                 return body
 
         elif self.status == httplib.NOT_FOUND:
-            if (not json_error) and ('error' in body):
+            if (not json_error) and ("error" in body):
                 (code, message) = self._get_error(body)
             else:
                 message = body
@@ -278,7 +286,7 @@ class GoogleResponse(JsonResponse):
             raise ResourceNotFoundError(message, self.status, code)
 
         elif self.status == httplib.BAD_REQUEST:
-            if (not json_error) and ('error' in body):
+            if (not json_error) and ("error" in body):
                 (code, message) = self._get_error(body)
             else:
                 message = body
@@ -286,7 +294,7 @@ class GoogleResponse(JsonResponse):
             raise InvalidRequestError(message, self.status, code)
 
         else:
-            if (not json_error) and ('error' in body):
+            if (not json_error) and ("error" in body):
                 (code, message) = self._get_error(body)
             else:
                 message = body
@@ -303,15 +311,22 @@ class GoogleBaseAuthConnection(ConnectionUserAndKey):
     Base class for Google Authentication.  Should be subclassed for specific
     types of authentication.
     """
+
     driver = GoogleBaseDriver
     responseCls = GoogleResponse
-    name = 'Google Auth'
-    host = 'accounts.google.com'
-    auth_path = '/o/oauth2/auth'
+    name = "Google Auth"
+    host = "accounts.google.com"
+    auth_path = "/o/oauth2/auth"
 
-    def __init__(self, user_id, key=None, scopes=None,
-                 redirect_uri='urn:ietf:wg:oauth:2.0:oob',
-                 login_hint=None, **kwargs):
+    def __init__(
+        self,
+        user_id,
+        key=None,
+        scopes=None,
+        redirect_uri="urn:ietf:wg:oauth:2.0:oob",
+        login_hint=None,
+        **kwargs,
+    ):
         """
         :param  user_id: The email address (for service accounts) or Client ID
                          (for installed apps) to be used for authentication.
@@ -347,8 +362,8 @@ class GoogleBaseAuthConnection(ConnectionUserAndKey):
         """
         Add defaults for 'Content-Type' and 'Host' headers.
         """
-        headers['Content-Type'] = "application/x-www-form-urlencoded"
-        headers['Host'] = self.host
+        headers["Content-Type"] = "application/x-www-form-urlencoded"
+        headers["Host"] = self.host
         return headers
 
     def _token_request(self, request_body):
@@ -364,16 +379,18 @@ class GoogleBaseAuthConnection(ConnectionUserAndKey):
         """
         data = urlencode(request_body)
         try:
-            response = self.request('/o/oauth2/token', method='POST',
-                                    data=data)
+            response = self.request("/o/oauth2/token", method="POST", data=data)
         except AttributeError:
-            raise GoogleAuthError('Invalid authorization response, please '
-                                  'check your credentials and time drift.')
+            raise GoogleAuthError(
+                "Invalid authorization response, please "
+                "check your credentials and time drift."
+            )
         token_info = response.object
-        if 'expires_in' in token_info:
+        if "expires_in" in token_info:
             expire_time = _utcnow() + datetime.timedelta(
-                seconds=token_info['expires_in'])
-            token_info['expire_time'] = _utc_timestamp(expire_time)
+                seconds=token_info["expires_in"]
+            )
+            token_info["expire_time"] = _utc_timestamp(expire_time)
         return token_info
 
     def refresh_token(self, token_info):
@@ -395,6 +412,7 @@ class GoogleBaseAuthConnection(ConnectionUserAndKey):
 
 class GoogleInstalledAppAuthConnection(GoogleBaseAuthConnection):
     """Authentication connection for "Installed Application" authentication."""
+
     def get_code(self):
         """
         Give the user a URL that they can visit to authenticate and obtain a
@@ -405,23 +423,25 @@ class GoogleInstalledAppAuthConnection(GoogleBaseAuthConnection):
         :return:  Code supplied by the user after authenticating
         :rtype:   ``str``
         """
-        auth_params = {'response_type': 'code',
-                       'client_id': self.user_id,
-                       'redirect_uri': self.redirect_uri,
-                       'scope': self.scopes,
-                       'state': 'Libcloud Request'}
+        auth_params = {
+            "response_type": "code",
+            "client_id": self.user_id,
+            "redirect_uri": self.redirect_uri,
+            "scope": self.scopes,
+            "state": "Libcloud Request",
+        }
         if self.login_hint:
-            auth_params['login_hint'] = self.login_hint
+            auth_params["login_hint"] = self.login_hint
 
         data = urlencode(auth_params)
 
-        url = 'https://%s%s?%s' % (self.host, self.auth_path, data)
-        print('\nPlease Go to the following URL and sign in:')
+        url = "https://%s%s?%s" % (self.host, self.auth_path, data)
+        print("\nPlease Go to the following URL and sign in:")
         print(url)
         if PY3:
-            code = input('Enter Code: ')
+            code = input("Enter Code: ")
         else:
-            code = raw_input('Enter Code: ')  # NOQA pylint: disable=undefined-variable
+            code = raw_input("Enter Code: ")  # NOQA pylint: disable=undefined-variable
         return code
 
     def get_new_token(self):
@@ -435,11 +455,13 @@ class GoogleInstalledAppAuthConnection(GoogleBaseAuthConnection):
         # Ask the user for a code
         code = self.get_code()
 
-        token_request = {'code': code,
-                         'client_id': self.user_id,
-                         'client_secret': self.key,
-                         'redirect_uri': self.redirect_uri,
-                         'grant_type': 'authorization_code'}
+        token_request = {
+            "code": code,
+            "client_id": self.user_id,
+            "client_secret": self.key,
+            "redirect_uri": self.redirect_uri,
+            "grant_type": "authorization_code",
+        }
 
         return self._token_request(token_request)
 
@@ -453,54 +475,94 @@ class GoogleInstalledAppAuthConnection(GoogleBaseAuthConnection):
         :return:  A dictionary containing updated token information.
         :rtype:   ``dict``
         """
-        if 'refresh_token' not in token_info:
+        if "refresh_token" not in token_info:
             return self.get_new_token()
-        refresh_request = {'refresh_token': token_info['refresh_token'],
-                           'client_id': self.user_id,
-                           'client_secret': self.key,
-                           'grant_type': 'refresh_token'}
+        refresh_request = {
+            "refresh_token": token_info["refresh_token"],
+            "client_id": self.user_id,
+            "client_secret": self.key,
+            "grant_type": "refresh_token",
+        }
 
         new_token = self._token_request(refresh_request)
-        if 'refresh_token' not in new_token:
-            new_token['refresh_token'] = token_info['refresh_token']
+        if "refresh_token" not in new_token:
+            new_token["refresh_token"] = token_info["refresh_token"]
         return new_token
 
 
 class GoogleServiceAcctAuthConnection(GoogleBaseAuthConnection):
     """Authentication class for "Service Account" authentication."""
+
     def __init__(self, user_id, key, *args, **kwargs):
         """
-        Check to see if cryptography is available, and convert key file path
-        into a key string if the key is in a file.
+        Check to see if cryptography is available, and convert PEM key file
+        into a key string, or extract the key from JSON object, string or
+        file.
 
         :param  user_id: Email address to be used for Service Account
                 authentication.
         :type   user_id: ``str``
 
-        :param  key: The RSA Key or path to file containing the key.
-        :type   key: ``str``
+        :param  key: The path to a PEM/JSON file containing the private RSA
+        key, or a str/dict containing the PEM/JSON.
+        :type   key: ``str`` or ``dict``
+
         """
         if SHA256 is None:
-            raise GoogleAuthError('cryptography library required for '
-                                  'Service Account Authentication.')
-        # Check to see if 'key' is a file and read the file if it is.
-        if key.find("PRIVATE KEY---") == -1:
-            # key is a file
-            keypath = os.path.expanduser(key)
-            is_file_path = os.path.exists(keypath) and os.path.isfile(keypath)
-            if not is_file_path:
-                raise ValueError("Missing (or not readable) key "
-                                 "file: '%s'" % key)
-            with open(keypath, 'r') as f:
-                contents = f.read()
-            try:
-                key = json.loads(contents)
-                key = key['private_key']
-            except ValueError:
-                key = contents
+            raise GoogleAuthError(
+                "cryptography library required for " "Service Account Authentication."
+            )
+
+        if isinstance(key, dict):
+            # if it's a dict, assume it's containing parsed JSON
+            key_content = key
+            key = None
+        else:
+            key_path = os.path.expanduser(key)
+            if os.path.exists(key_path) and os.path.isfile(key_path):
+                # Assume it's a file and read it
+                try:
+                    with open(key_path, "r") as f:
+                        key_content = f.read()
+                except IOError:
+                    raise GoogleAuthError(
+                        "Missing (or unreadable) key " "file: '%s'" % key
+                    )
+            else:
+                # assume it's a PEM str or serialized JSON str
+                key_content = key
+                key = None
+
+        try:
+            key_content = json.loads(key_content)
+            # check if serialized JSON given
+        except TypeError:
+            # already a dict
+            pass
+        except ValueError:
+            # it's been a PEM string all along
+            pass
+        finally:
+            if "private_key" in key_content:
+                key = key_content["private_key"]
+            else:
+                key = key_content
+
+        try:
+            # check if the key is actually a PEM encoded private key
+            serialization.load_pem_private_key(
+                b(key), password=None, backend=default_backend()
+            )
+        except ValueError as e:
+            raise GoogleAuthError("Unable to decode provided PEM key: %s" % e)
+        except TypeError as e:
+            raise GoogleAuthError("Unable to decode provided PEM key: %s" % e)
+        except exceptions.UnsupportedAlgorithm as e:
+            raise GoogleAuthError("Unable to decode provided PEM key: %s" % e)
 
         super(GoogleServiceAcctAuthConnection, self).__init__(
-            user_id, key, *args, **kwargs)
+            user_id, key, *args, **kwargs
+        )
 
     def get_new_token(self):
         """
@@ -510,36 +572,34 @@ class GoogleServiceAcctAuthConnection(GoogleBaseAuthConnection):
         :rtype:   ``dict``
         """
         # The header is always the same
-        header = {'alg': 'RS256', 'typ': 'JWT'}
+        header = {"alg": "RS256", "typ": "JWT"}
         header_enc = base64.urlsafe_b64encode(b(json.dumps(header)))
 
         # Construct a claim set
-        claim_set = {'iss': self.user_id,
-                     'scope': self.scopes,
-                     'aud': 'https://accounts.google.com/o/oauth2/token',
-                     'exp': int(time.time()) + 3600,
-                     'iat': int(time.time())}
+        claim_set = {
+            "iss": self.user_id,
+            "scope": self.scopes,
+            "aud": "https://accounts.google.com/o/oauth2/token",
+            "exp": int(time.time()) + 3600,
+            "iat": int(time.time()),
+        }
         claim_set_enc = base64.urlsafe_b64encode(b(json.dumps(claim_set)))
 
         # The message contains both the header and claim set
-        message = b'.'.join((header_enc, claim_set_enc))
+        message = b".".join((header_enc, claim_set_enc))
         # Then the message is signed using the key supplied
         key = serialization.load_pem_private_key(
-            b(self.key),
-            password=None,
-            backend=default_backend()
-        )
-        signature = key.sign(
-            data=b(message),
-            padding=PKCS1v15(),
-            algorithm=SHA256()
+            b(self.key), password=None, backend=default_backend()
         )
+        signature = key.sign(data=b(message), padding=PKCS1v15(), algorithm=SHA256())
         signature = base64.urlsafe_b64encode(signature)
 
         # Finally the message and signature are sent to get a token
-        jwt = b'.'.join((message, signature))
-        request = {'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
-                   'assertion': jwt}
+        jwt = b".".join((message, signature))
+        request = {
+            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
+            "assertion": jwt,
+        }
 
         return self._token_request(request)
 
@@ -548,6 +608,7 @@ class GoogleGCEServiceAcctAuthConnection(GoogleBaseAuthConnection):
     """Authentication class for self-authentication when used with a GCE
     instance that supports serviceAccounts.
     """
+
     def get_new_token(self):
         """
         Get a new token from the internal metadata service.
@@ -555,19 +616,22 @@ class GoogleGCEServiceAcctAuthConnection(GoogleBaseAuthConnection):
         :return:  Dictionary containing token information
         :rtype:   ``dict``
         """
-        path = '/instance/service-accounts/default/token'
+        path = "/instance/service-accounts/default/token"
         http_code, http_reason, token_info = _get_gce_metadata(path)
         if http_code == httplib.NOT_FOUND:
-            raise ValueError("Service Accounts are not enabled for this "
-                             "GCE instance.")
+            raise ValueError(
+                "Service Accounts are not enabled for this " "GCE instance."
+            )
         if http_code != httplib.OK:
-            raise ValueError("Internal GCE Authorization failed: "
-                             "'%s'" % str(http_reason))
+            raise ValueError(
+                "Internal GCE Authorization failed: " "'%s'" % str(http_reason)
+            )
         token_info = json.loads(token_info)
-        if 'expires_in' in token_info:
+        if "expires_in" in token_info:
             expire_time = _utcnow() + datetime.timedelta(
-                seconds=token_info['expires_in'])
-            token_info['expire_time'] = _utc_timestamp(expire_time)
+                seconds=token_info["expires_in"]
+            )
+            token_info["expire_time"] = _utc_timestamp(expire_time)
         return token_info
 
 
@@ -578,10 +642,11 @@ class GoogleAuthType(object):
     GCE (Auth from a GCE instance with service account enabled)
     GCS_S3 (Cloud Storage S3 interoperability authentication)
     """
-    SA = 'SA'
-    IA = 'IA'
-    GCE = 'GCE'
-    GCS_S3 = 'GCS_S3'
+
+    SA = "SA"
+    IA = "IA"
+    GCE = "GCE"
+    GCS_S3 = "GCS_S3"
 
     ALL_TYPES = [SA, IA, GCE, GCS_S3]
     OAUTH2_TYPES = [SA, IA, GCE]
@@ -592,22 +657,36 @@ class GoogleAuthType(object):
             return cls.SA
         elif cls._is_gcs_s3(user_id):
             return cls.GCS_S3
+        elif cls._is_installed_application(user_id):
+            # NOTE: This should be before "_is_gce()" call so we avoid
+            # querying GCE metadata service if that's not necessary
+            return cls.IA
         elif cls._is_gce():
             return cls.GCE
         else:
+            # TODO: It's probably safe to throw here, but we return cls.IA
+            # for backward compatibility reasons
             return cls.IA
 
     @classmethod
     def is_oauth2(cls, auth_type):
         return auth_type in cls.OAUTH2_TYPES
 
+    @staticmethod
+    def _is_installed_application(user_id):
+        return user_id.endswith("apps.googleusercontent.com")
+
     @staticmethod
     def _is_gce():
         """
         Checks if we can access the GCE metadata server.
         Mocked in libcloud.test.common.google.GoogleTestCase.
         """
-        http_code, http_reason, body = _get_gce_metadata()
+        # When using oAuth credentials we check for metadata server first, so
+        # if server is unavailable and we retry many times before timing out,
+        # this will slow down the driver instantiation when retrying failed
+        # requests is enabled globally.
+        http_code, http_reason, body = _get_gce_metadata(retry_failed=False)
         if http_code == httplib.OK and body:
             return True
         return False
@@ -617,51 +696,54 @@ class GoogleAuthType(object):
         """
         Checks S3 key format: alphanumeric chars starting with GOOG.
         """
-        return user_id.startswith('GOOG')
+        return user_id.startswith("GOOG")
 
     @staticmethod
     def _is_sa(user_id):
-        return user_id.endswith('.gserviceaccount.com')
+        return user_id.endswith(".gserviceaccount.com")
 
 
 class GoogleOAuth2Credential(object):
-    default_credential_file = '~/.google_libcloud_auth'
+    default_credential_file = "~/.google_libcloud_auth"
 
-    def __init__(self, user_id, key, auth_type=None, credential_file=None,
-                 scopes=None, **kwargs):
+    def __init__(
+        self, user_id, key, auth_type=None, credential_file=None, scopes=None, **kwargs
+    ):
         self.auth_type = auth_type or GoogleAuthType.guess_type(user_id)
         if self.auth_type not in GoogleAuthType.ALL_TYPES:
-            raise GoogleAuthError('Invalid auth type: %s' % self.auth_type)
+            raise GoogleAuthError("Invalid auth type: %s" % self.auth_type)
         if not GoogleAuthType.is_oauth2(self.auth_type):
-            raise GoogleAuthError(('Auth type %s cannot be used with OAuth2' %
-                                   self.auth_type))
+            raise GoogleAuthError(
+                ("Auth type %s cannot be used with OAuth2" % self.auth_type)
+            )
         self.user_id = user_id
         self.key = key
 
-        default_credential_file = '.'.join([self.default_credential_file,
-                                            user_id])
+        default_credential_file = ".".join([self.default_credential_file, user_id])
         self.credential_file = credential_file or default_credential_file
         # Default scopes to read/write for compute, storage, and dns.
         self.scopes = scopes or [
-            'https://www.googleapis.com/auth/compute',
-            'https://www.googleapis.com/auth/devstorage.full_control',
-            'https://www.googleapis.com/auth/ndev.clouddns.readwrite',
+            "https://www.googleapis.com/auth/compute",
+            "https://www.googleapis.com/auth/devstorage.full_control",
+            "https://www.googleapis.com/auth/ndev.clouddns.readwrite",
         ]
 
         self.token = self._get_token_from_file()
 
         if self.auth_type == GoogleAuthType.GCE:
             self.oauth2_conn = GoogleGCEServiceAcctAuthConnection(
-                self.user_id, self.scopes, **kwargs)
+                self.user_id, self.scopes, **kwargs
+            )
         elif self.auth_type == GoogleAuthType.SA:
             self.oauth2_conn = GoogleServiceAcctAuthConnection(
-                self.user_id, self.key, self.scopes, **kwargs)
+                self.user_id, self.key, self.scopes, **kwargs
+            )
         elif self.auth_type == GoogleAuthType.IA:
             self.oauth2_conn = GoogleInstalledAppAuthConnection(
-                self.user_id, self.key, self.scopes, **kwargs)
+                self.user_id, self.key, self.scopes, **kwargs
+            )
         else:
-            raise GoogleAuthError('Invalid auth_type: %s' %
-                                  str(self.auth_type))
+            raise GoogleAuthError("Invalid auth_type: %s" % str(self.auth_type))
 
         if self.token is None:
             self.token = self.oauth2_conn.get_new_token()
@@ -671,11 +753,11 @@ class GoogleOAuth2Credential(object):
     def access_token(self):
         if self.token_expire_utc_datetime < _utcnow():
             self._refresh_token()
-        return self.token['access_token']
+        return self.token["access_token"]
 
     @property
     def token_expire_utc_datetime(self):
-        return _from_utc_timestamp(self.token['expire_time'])
+        return _from_utc_timestamp(self.token["expire_time"])
 
     def _refresh_token(self):
         self.token = self.oauth2_conn.refresh_token(self.token)
@@ -693,14 +775,15 @@ class GoogleOAuth2Credential(object):
         filename = os.path.realpath(os.path.expanduser(self.credential_file))
 
         try:
-            with open(filename, 'r') as f:
+            with open(filename, "r") as f:
                 data = f.read()
             token = json.loads(data)
         except (IOError, ValueError) as e:
             # Note: File related errors (IOError) and errors related to json
             # parsing of the data (ValueError) are not fatal.
-            LOG.info('Failed to read cached auth token from file "%s": %s',
-                     filename, str(e))
+            LOG.info(
+                'Failed to read cached auth token from file "%s": %s', filename, str(e)
+            )
 
         return token
 
@@ -715,15 +798,13 @@ class GoogleOAuth2Credential(object):
         try:
             data = json.dumps(self.token)
             write_flags = os.O_CREAT | os.O_WRONLY | os.O_TRUNC
-            with os.fdopen(os.open(filename, write_flags,
-                                   int('600', 8)), 'w') as f:
+            with os.fdopen(os.open(filename, write_flags, int("600", 8)), "w") as f:
                 f.write(data)
         except Exception as e:
             # Note: Failure to write (cache) token in a file is not fatal. It
             # simply means degraded performance since we will need to acquire a
             # new token each time script runs.
-            LOG.info('Failed to write auth token to file "%s": %s',
-                     filename, str(e))
+            LOG.info('Failed to write auth token to file "%s": %s', filename, str(e))
 
 
 class GoogleBaseConnection(ConnectionUserAndKey, PollingConnection):
@@ -731,12 +812,19 @@ class GoogleBaseConnection(ConnectionUserAndKey, PollingConnection):
 
     driver = GoogleBaseDriver
     responseCls = GoogleResponse
-    host = 'www.googleapis.com'
+    host = "www.googleapis.com"
     poll_interval = 2.0
     timeout = 180
 
-    def __init__(self, user_id, key=None, auth_type=None,
-                 credential_file=None, scopes=None, **kwargs):
+    def __init__(
+        self,
+        user_id,
+        key=None,
+        auth_type=None,
+        credential_file=None,
+        scopes=None,
+        **kwargs,
+    ):
         """
         Determine authentication type, set up appropriate authentication
         connection and get initial authentication information.
@@ -768,19 +856,23 @@ class GoogleBaseConnection(ConnectionUserAndKey, PollingConnection):
         super(GoogleBaseConnection, self).__init__(user_id, key, **kwargs)
 
         self.oauth2_credential = GoogleOAuth2Credential(
-            user_id, key, auth_type, credential_file, scopes, **kwargs)
+            user_id, key, auth_type, credential_file, scopes, **kwargs
+        )
 
-        python_ver = '%s.%s.%s' % (sys.version_info[0], sys.version_info[1],
-                                   sys.version_info[2])
-        ver_platform = 'Python %s/%s' % (python_ver, sys.platform)
+        python_ver = "%s.%s.%s" % (
+            sys.version_info[0],
+            sys.version_info[1],
+            sys.version_info[2],
+        )
+        ver_platform = "Python %s/%s" % (python_ver, sys.platform)
         self.user_agent_append(ver_platform)
 
     def add_default_headers(self, headers):
         """
         @inherits: :class:`Connection.add_default_headers`
         """
-        headers['Content-Type'] = 'application/json'
-        headers['Host'] = self.host
+        headers["Content-Type"] = "application/json"
+        headers["Host"] = self.host
         return headers
 
     def pre_connect_hook(self, params, headers):
@@ -790,8 +882,7 @@ class GoogleBaseConnection(ConnectionUserAndKey, PollingConnection):
 
         @inherits: :class:`Connection.pre_connect_hook`
         """
-        headers['Authorization'] = ('Bearer ' +
-                                    self.oauth2_credential.access_token)
+        headers["Authorization"] = "Bearer " + self.oauth2_credential.access_token
         return params, headers
 
     def encode_data(self, data):
@@ -808,8 +899,7 @@ class GoogleBaseConnection(ConnectionUserAndKey, PollingConnection):
         tries = 0
         while tries < (retries - 1):
             try:
-                return super(GoogleBaseConnection, self).request(
-                    *args, **kwargs)
+                return super(GoogleBaseConnection, self).request(*args, **kwargs)
             except socket.error as e:
                 if e.errno == errno.ECONNRESET:
                     tries = tries + 1
@@ -828,7 +918,7 @@ class GoogleBaseConnection(ConnectionUserAndKey, PollingConnection):
         :return:  True if complete, False otherwise
         :rtype:   ``bool``
         """
-        if response.object['status'] == 'DONE':
+        if response.object["status"] == "DONE":
             return True
         else:
             return False
@@ -837,7 +927,7 @@ class GoogleBaseConnection(ConnectionUserAndKey, PollingConnection):
         """
         @inherits: :class:`PollingConnection.get_poll_request_kwargs`
         """
-        return {'action': response.object['selfLink']}
+        return {"action": response.object["selfLink"]}
 
     def morph_action_hook(self, action):
         """
@@ -854,9 +944,9 @@ class GoogleBaseConnection(ConnectionUserAndKey, PollingConnection):
         :return:  The modified request based on the action
         :rtype:   ``str``
         """
-        if action.startswith('https://'):
+        if action.startswith("https://"):
             u = urlparse.urlsplit(action)
-            request = urlparse.urlunsplit(('', '', u[2], u[3], u[4]))
+            request = urlparse.urlunsplit(("", "", u[2], u[3], u[4]))
         else:
             request = self.request_path + action
         return request
